import yapecs
from pathlib import Path

import torbi


###############################################################################
# Entry point
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = yapecs.ArgumentParser(
        description='Synthesize speech from features')
    parser.add_argument(
        '--input_files',
        type=Path,
        nargs='+',
        required=True,
        help='Time-varying categorical distribution files')
    parser.add_argument(
        '--output_files',
        type=Path,
        nargs='+',
        required=True,
        help='Files to save decoded indices')
    parser.add_argument(
        '--transition_file',
        type=Path,
        help='Categorical transition matrix file; defaults to uniform')
    parser.add_argument(
        '--initial_file',
        type=Path,
        help='Categorical initial distribution file; defaults to uniform')
    parser.add_argument(
        '--log_probs',
        action='store_true',
        help='Whether inputs are in (natural) log space')
    parser.add_argument(
        '--gpu',
        type=int,
        help='GPU index to use for decoding. Defaults to CPU.')
    parser.add_argument(
        '--num_threads',
        type=int,
        default = 1,
        help='The number of threads to use for parellelized CPU decoding'
    )
    return parser.parse_known_args()[0]


torbi.from_files_to_files(**vars(parse_args()))



from typing import List

import torch
import torbi


###############################################################################
# Chunked Viterbi decoding; quickly and accurately batch process long sequences
###############################################################################


def chunk(
    observation: torch.Tensor,
    min_chunk_size: int = torbi.MIN_CHUNK_SIZE,
    entropy_threshold: float = torbi.ENTROPY_THRESHOLD
) -> List:
    """Chunk observations based on points of low entropy

    Arguments
        observation
            Time-varying categorical distribution
            shape=(frames, states)
        min_chunk_size
            Minimum chunk size in frames
        entropy_threshold
            Threshold for entropy to allow splitting

    Returns
        chunks
            List of chunked sequence data
    """
    start = 0
    chunks = []

    # Get split points
    for split_point in split(
        observation,
        min_chunk_size=min_chunk_size,
        entropy_threshold=entropy_threshold
    ):

        # Chunk observations
        chunks.append(observation[start:split_point])
        start = split_point

     # Get last chunk
    chunks.append(observation[start:])

    return chunks


###############################################################################
# Utilities
###############################################################################


def split(
    observation,
    min_chunk_size=torbi.MIN_CHUNK_SIZE,
    entropy_threshold=torbi.ENTROPY_THRESHOLD
) -> List[int]:
    """Find split points of minimum entropy"""
    observation = observation.T

    # Find low-entropy time frames
    candidates = entropy(observation) < entropy_threshold

    # Find adjacent low-entropy time frames
    split_points = []
    i = min_chunk_size
    while i < observation.shape[-1]:
        if candidates[i] and candidates[i - 1]:
            split_points.append(i)
            i += min_chunk_size
        else:
            i += 1

    return split_points


def entropy(observation: torch.Tensor) -> torch.Tensor:
    """Compute framewise entropy of a sequence of categorical distributions"""
    return -(
        (torch.exp(observation) * observation).sum(dim=0) /
        torch.log(torch.tensor(observation.shape[0])))



import torch
import torbi


###############################################################################
# Dataset
###############################################################################


class Dataset(torch.utils.data.Dataset):

    def __init__(self, input_files):
        self.input_files = input_files

    def __getitem__(self, index):
        """Retrieve the indexth item"""
        input_file = self.input_files[index]

        observation = torch.load(input_file, map_location='cpu')

        # Maybe chunk observations
        if torbi.MIN_CHUNK_SIZE is not None:
            observation = torbi.chunk(observation)

        return observation, input_file

    def __len__(self):
        """Length of the dataset"""
        return len(self.input_files)



import torch
import torchutil

import torbi


###############################################################################
# Aggregate metric
###############################################################################


class Metrics:

    def __init__(self):
        self.rpas = [RPA(k) for k in torbi.PITCH_ERROR_THRESHOLDS]

    def __call__(self):
        return {'rpa': {str(rpa.threshold): rpa() for rpa in self.rpas}}

    def update(self, predicted, target):
        for rpa in self.rpas:
            rpa.update(predicted, target)

    def reset(self):
        for rpa in self.rpas:
            rpa.reset()


###############################################################################
# Individual metrics
###############################################################################


class RPA(torchutil.metrics.Average):

    def __init__(self, threshold):
        super().__init__()
        self.threshold = threshold

    def update(self, predicted, target):
        super().update(
            torch.abs(predicted - target) <= self.threshold,
            predicted.numel())



import torch
from typing import Optional


def decode(
    observation: torch.Tensor,
    batch_frames: torch.Tensor,
    transition: torch.Tensor,
    initial: torch.Tensor,
    num_threads: int = 0,
) -> torch.Tensor:
    """Decode a time-varying categorical distribution

    Args:
        observation: :math:`(N, T, S)` or :math:`(T, S)`
            where `S = the number of states`,
            `T = the length of the sequence`,
            and `N = batch size`.
            Time-varying categorical distribution
        batch_frames :math:`(N)`
            Sequence length of each batch item
        transition :math:`(S, S)`
            Categorical transition matrix
        initial :math:`(S)`
            Categorical initial distribution

    Return:
        indices: :math:`(N, T)`
            The decoded bin indices

    Example::

            >>> observation = torch.tensor([[
            >>>     [0.25, 0.5, 0.25],
            >>>     [0.25, 0.25, 0.5],
            >>>     [0.33, 0.33, 0.33]
            >>> ]])
            >>> batch_frames = torch.tensor([3])
            >>> transition = torch.tensor([
            >>>     [0.5, 0.25, 0.25],
            >>>     [0.33, 0.34, 0.33],
            >>>     [0.25, 0.25, 0.5]
            >>> ])
            >>> initial = torch.tensor([0.4, 0.35, 0.25])
            >>> bins = viterbi.decode(
            >>>     observation,
            >>>     batch_frames,
            >>>     transition,
            >>>     initial)
    """
    if observation.device.type == 'cpu':
        torch.set_num_threads(num_threads)
    return torch.ops.torbi.viterbi_decode(observation, batch_frames, transition, initial)



import torch


###############################################################################
# Dataloader
###############################################################################


def collate(batch):
    observations, input_files = zip(*batch)

    # Handle chunking
    if isinstance(observations[0], list):
        batch_chunks = [len(obs) for obs in observations]
        observations = sum(observations, [])
    else:
        batch_chunks = [1] * len(observations)
    batch_frames = torch.tensor([obs.shape[0] for obs in observations])

    batch = len(observations)
    if batch == 0:
        raise ValueError('batch must contain at least 1 item')

    max_frames = max(observation.shape[0] for observation in observations)

    observation = torch.zeros(
        (batch, max_frames, observations[0].shape[-1]),
        dtype=observations[0].dtype)

    for i, obs in enumerate(observations):
        observation[i, :obs.shape[0]] = obs

    return observation, batch_frames, batch_chunks, input_files


def separate(indices, batch_chunks, batch_frames):
    start = 0
    separated = []
    for chunks in batch_chunks:
        frames = batch_frames[start:start+chunks]
        separated.append(
            torch.cat([
                indices[start + i, :frames[i]] for i in range(0, chunks)]))
        start += chunks
    return separated



import torch
import torchutil

import torbi


###############################################################################
# Preprocess datasets
###############################################################################


@torchutil.notify('preprocess')
def datasets(datasets, gpu=None):
    """Preprocess a dataset"""
    try:
        import penn
    except ImportError:
        raise ImportError("penn is required for evaluation. Please install torbi with `torbi[evaluate]`")

    for dataset in datasets:

        # Get cache directory
        directory = torbi.CACHE_DIR / dataset

        # Get files
        audio_files = sorted(list(directory.rglob('*.wav')))

        # Preprocess pitch posteriorgrams
        for audio_file in audio_files:
            logits = []

            # Load audio file
            audio = penn.load.audio(audio_file)

            # Preprocess audio
            for frames in penn.preprocess(audio, center='half-hop'):

                # Copy to device
                frames = frames.to('cpu' if gpu is None else f'cuda:{gpu}')

                # Infer
                logits = penn.infer(frames).detach()

            # Concatenate results
            if isinstance(logits, torch.Tensor):
                logits = [logits]
            logits = torch.cat(logits, dim=0).squeeze(2)

            # Normalize
            logits = torch.nn.functional.log_softmax(logits, dim=1)

            # Save to cache
            torch.save(logits, audio_file.with_suffix('.pt'))



import torch.utils.cpp_extension
from pathlib import Path

lib = torch.utils.cpp_extension.load(
    name='viterbi_mps',
    sources=[Path(__file__).parent / 'viterbi.mm'],
    extra_cflags=['-std=c++17'],
    is_python_module=False,
    # verbose=True
)


import json
import random

import torbi


###############################################################################
# Partition
###############################################################################


def datasets(datasets):
    """Partition datasets and save to disk"""
    for dataset in datasets:
        random.seed(torbi.RANDOM_SEED)

        # Get stems
        directory = torbi.CACHE_DIR / dataset
        stems = [
            f'{file.parent.name}/{file.stem}'
            for file in directory.rglob('*.wav')]

        # Shuffle
        random.shuffle(stems)

        # Slice
        stems = stems[:torbi.EVALUATION_SAMPLES]

        # Save to disk
        file = torbi.PARTITION_DIR / f'{dataset}.json'
        file.parent.mkdir(exist_ok=True, parents=True)
        with open(file, 'w') as file:
            json.dump(stems, file, indent=4)



###############################################################################
# Configuration
###############################################################################


# Default configuration parameters to be modified
from .config import defaults

# Modify configuration
import yapecs
yapecs.configure('torbi', defaults)

# Import configuration parameters
from .config.defaults import *
import torbi
del torbi.defaults # remove unnecessary module
from .config.static import *


###############################################################################
# Module imports
###############################################################################

import torch
# load the library .so
# taken from https://docs.pytorch.org/tutorials/advanced/cpp_custom_ops.html
from pathlib import Path
import platform
if platform.system() == "Windows":
    pyd_files = list(Path(__file__).parent.glob("_C*.pyd"))
    assert (
        len(pyd_files) == 1
    ), f"Expected one _C*.pyd file, found {len(pyd_files)}"
    torch.ops.load_library(pyd_files[0])
else:
    so_files = list(Path(__file__).parent.glob("_C*.so"))
    assert (
        len(so_files) == 1
    ), f"Expected one _C*.so file, found {len(so_files)}"
    torch.ops.load_library(so_files[0])

if torch.backends.mps.is_available():
    import warnings
    try:
        import ninja
        from . import mps
    except ImportError:
        warnings.warn("ninja is not installed, so the mps backend cannot be loaded.")
    except RuntimeError as e:
        warnings.warn("Could not compile mps backend:", e)

from .viterbi import decode
from .core import *
from .chunk import chunk
from . import data
from . import evaluate
from . import partition
from . import reference



import yapecs

import torbi


###############################################################################
# Partition datasets
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = yapecs.ArgumentParser(description='Partition datasets')
    parser.add_argument(
        '--datasets',
        default=torbi.DATASETS,
        nargs='+',
        help='The datasets to partition')
    return parser.parse_args()


torbi.partition.datasets(**vars(parse_args()))



import torch
import numpy as np
import time

import torbi

def test_decode():
    """Viterbi decoding test"""
    observation = torch.tensor([
        [0.25, 0.5, 0.25],
        [0.25, 0.25, 0.5],
        [0.33, 0.33, 0.33]
    ]).unsqueeze(dim=0)
    transition = torch.tensor([
        [0.5, 0.25, 0.25],
        [0.33, 0.34, 0.33],
        [0.25, 0.25, 0.5]
    ])
    initial = torch.tensor([0.4, 0.35, 0.25])
    bins = torbi.from_probabilities(
        observation=observation,
        transition=transition,
        initial=initial,
        log_probs=False)
    assert (bins == torch.tensor([1, 2, 2])).all()

def test_cuda():
    """Viterbi decoding test with CUDA"""
    observation = torch.tensor([
        [0.25, 0.5, 0.25],
        [0.25, 0.25, 0.5],
        [0.33, 0.33, 0.33]
    ]).unsqueeze(dim=0).to('cuda:0')
    transition = torch.tensor([
        [0.5, 0.25, 0.25],
        [0.33, 0.34, 0.33],
        [0.25, 0.25, 0.5]
    ]).to('cuda:0')
    initial = torch.tensor([0.4, 0.35, 0.25]).to('cuda:0')
    bins = torbi.from_probabilities(
        observation=observation,
        transition=transition,
        initial=initial,
        log_probs=False,
        gpu=0)
    assert (bins == torch.tensor([1, 2, 2]).to('cuda:0')).all()



import shutil
from pathlib import Path

import torch
import torchutil
import torchaudio
import zipfile
import json

import torbi


###############################################################################
# Download datasets
###############################################################################


@torchutil.notify('download')
def datasets(datasets):
    """Download datasets"""
    # Download and format daps dataset
    if 'daps' in datasets:
        daps()
    if 'vctk' in datasets:
        vctk()


def daps():
    """Download daps dataset"""
    import torchaudio

    # Download tarball
    torchutil.download.targz(
        'https://zenodo.org/record/4783456/files/daps-segmented.tar.gz?download=1',
        torbi.DATA_DIR)

    # Delete previous directory
    shutil.rmtree(torbi.DATA_DIR / 'daps', ignore_errors=True)

    # Rename directory
    data_directory = torbi.DATA_DIR / 'daps'
    shutil.move(
        torbi.DATA_DIR / 'daps-segmented',
        data_directory)

    # Get audio files
    audio_files = sorted(
        [path.resolve() for path in data_directory.rglob('*.wav')])
    text_files = [file.with_suffix('.txt') for file in audio_files]

    # Write audio to cache
    speaker_count = {}
    cache_directory = torbi.CACHE_DIR / 'daps'
    cache_directory.mkdir(exist_ok=True, parents=True)
    with torchutil.paths.chdir(cache_directory):

        # Iterate over files
        for audio_file, text_file in torchutil.iterator(
            zip(audio_files, text_files),
            'Formatting daps',
            total=len(audio_files)
        ):

            # Get speaker ID
            speaker = Path(audio_file.stem.split('_')[0])
            if speaker not in speaker_count:

                # Each entry is (index, count)
                speaker_count[speaker] = [len(speaker_count), 0]

            # Update speaker and get current entry
            speaker_count[speaker][1] += 1
            index, count = speaker_count[speaker]

            # Load audio
            audio, sample_rate = torchaudio.load(audio_file)

            # If audio is too quiet, increase the volume
            maximum = torch.abs(audio).max()
            if maximum < .35:
                audio *= .35 / maximum

            # Save at original sampling rate
            speaker_directory = cache_directory / f'{index:04d}'
            speaker_directory.mkdir(exist_ok=True, parents=True)
            output_file = Path(f'{count:06d}.wav')
            torchaudio.save(
                speaker_directory / output_file,
                audio,
                sample_rate)
            shutil.copyfile(
                text_file,
                (speaker_directory / output_file).with_suffix('.txt'))

def vctk():
    """Download vctk dataset"""
    directory = torbi.DATA_DIR / 'vctk'
    directory.mkdir(exist_ok=True, parents=True)
    torchutil.download.zip(
        'https://datashare.ed.ac.uk/download/DS_10283_3443.zip',
        directory)

    # Unzip
    for file in directory.glob('*.zip'):
        with zipfile.ZipFile(file) as zfile:
            zfile.extractall(directory)

    # File locations
    audio_directory = directory / 'wav48_silence_trimmed'

    # Get source files
    audio_files = sorted(list(audio_directory.rglob('*.flac')))
    text_files = [vctk_audio_file_to_text_file(file) for file in audio_files]

    # If the text file doesn't exist, remove corresponding audio file
    text_files = [file for file in text_files if file.exists()]
    audio_files = [
        file for file in audio_files
        if vctk_audio_file_to_text_file(file).exists()]

    # Write audio to cache
    speaker_count = {}
    correspondence = {}
    output_directory = torbi.CACHE_DIR / 'vctk'
    output_directory.mkdir(exist_ok=True, parents=True)
    with torchutil.paths.chdir(output_directory):

        # Iterate over files
        for audio_file, text_file in torchutil.iterator(
            zip(audio_files, text_files),
            'Formatting vctk',
            total=len(audio_files)
        ):

            # Get speaker ID
            speaker = Path(audio_file.stem.split('_')[0])
            if speaker not in speaker_count:

                # Each entry is (index, count)
                speaker_count[speaker] = [len(speaker_count), 0]

            # Update speaker and get current entry
            speaker_count[speaker][1] += 1
            index, count = speaker_count[speaker]

            # Load audio
            audio, sample_rate = torchaudio.load(audio_file)

            # If audio is too quiet, increase the volume
            maximum = torch.abs(audio).max()
            if maximum < .35:
                audio *= .35 / maximum

            # Save at original sampling rate
            speaker_directory = output_directory / f'{index:04d}'
            speaker_directory.mkdir(exist_ok=True, parents=True)
            output_file = Path(f'{count:06d}.wav')
            torchaudio.save(
                speaker_directory / output_file,
                audio,
                sample_rate)
            shutil.copyfile(
                text_file,
                (speaker_directory / output_file).with_suffix('.txt'))

            # Save at system sample rate
            audio = resample(audio, sample_rate)
            torchaudio.save(
                speaker_directory / f'{output_file.stem}-100.wav',
                audio,
                torbi.SAMPLE_RATE)

            # Save file stem correpondence
            correspondence[f'{index:04d}/{output_file.stem}'] = audio_file.stem
        with open('correspondence.json', 'w') as file:
            json.dump(correspondence, file)


###############################################################################
# Utilities
###############################################################################


def resample(audio, sample_rate):
    """Resample audio to ProMoNet sample rate"""
    # Cache resampling filter
    key = str(sample_rate)
    if not hasattr(resample, key):
        setattr(
            resample,
            key,
            torchaudio.transforms.Resample(sample_rate, torbi.SAMPLE_RATE))

    # Resample
    return getattr(resample, key)(audio)


def vctk_audio_file_to_text_file(audio_file):
    """Convert audio file to corresponding text file"""
    text_directory = torbi.DATA_DIR / 'vctk' / 'txt'
    return (
        text_directory /
        audio_file.parent.name /
        f'{audio_file.stem[:-5]}.txt')


def vctk_text_file_to_audio_file(text_file):
    """Convert audio file to corresponding text file"""
    audio_directory = torbi.DATA_DIR / 'vctk' / 'wav48_silence_trimmed'
    return (
        audio_directory /
        text_file.parent.name /
        f'{text_file.stem}.flac')



import json

import torch
import torchutil

import torbi


###############################################################################
# Evaluate
###############################################################################


def datasets(datasets, gpu=None, num_threads=1):
    """Evaluate Viterbi decoding methods"""
    # Cache transition matrix

    try:
        import penn
    except ImportError:
        raise ImportError("penn is required for evaluation. Please install torbi with `torbi[evaluate]`")

    if not torbi.PITCH_TRANSITION_MATRIX.exists():
        xx, yy = torch.meshgrid(
            torch.arange(penn.PITCH_BINS),
            torch.arange(penn.PITCH_BINS),
            indexing='ij')
        bins_per_octave = penn.OCTAVE / penn.CENTS_PER_BIN
        max_octaves_per_frame = \
            penn.MAX_OCTAVES_PER_SECOND * penn.HOPSIZE / penn.SAMPLE_RATE
        max_bins_per_frame = max_octaves_per_frame * bins_per_octave + 1
        transition = torch.clip(max_bins_per_frame - (xx - yy).abs(), 0)
        transition = transition / transition.sum(dim=1, keepdims=True)
        torch.save(transition, torbi.PITCH_TRANSITION_MATRIX)

    results = {}
    for dataset in datasets:

        # Reset benchmarking
        torchutil.time.reset()

        # Get evaluation stems
        with open(torbi.PARTITION_DIR / f'{dataset}.json') as file:
            stems = json.load(file)

        # Get input files
        input_files = [
            torbi.CACHE_DIR / dataset / f'{stem}.pt' for stem in stems]

        # get transition file
        transition_file = torbi.PITCH_TRANSITION_MATRIX

        # Get location of reference outputs
        if torbi.COMPARE_WITH_REFERENCE:
            reference_files = [
                torbi.EVAL_DIR /
                dataset /
                'reference' /
                f'{stem}.pt' for stem in stems]

            # Create parent directories
            (torbi.EVAL_DIR / dataset / 'reference').mkdir(
                parents=True,
                exist_ok=True)
            for file in reference_files:
                file.parent.mkdir(exist_ok=True)

            # Run reference Librosa implementation if we haven't yet
            if not all(file.exists() for file in reference_files):
                torbi.reference.from_files_to_files(
                    input_files,
                    reference_files,
                    transition_file=transition_file,
                    log_probs=True,
                    num_threads=num_threads)

        else: # Compare with non-chunked
            reference_files = [
                torbi.EVAL_DIR /
                dataset /
                'torbi' /
                f'{stem}.pt' for stem in stems]

        # Get location to save output
        output_files = [
            torbi.EVAL_DIR /
            dataset /
            torbi.CONFIG /
            f'{stem}.pt' for stem in stems]

        # Create parent directories
        (torbi.EVAL_DIR / dataset / torbi.CONFIG).mkdir(parents=True, exist_ok=True)
        for file in output_files:
            file.parent.mkdir(exist_ok=True)

        # Run Viterbi decoding
        torbi.from_files_to_files(
            input_files,
            output_files,
            transition_file=transition_file,
            log_probs=True,
            gpu=gpu,
            num_threads=num_threads)

        # Initialize metrics
        metrics = torbi.evaluate.Metrics()

        # Evaluate
        for predicted_file, target_file in zip(output_files, reference_files):
            predicted = torch.load(predicted_file, map_location='cpu')
            target = torch.load(target_file, map_location='cpu')
            metrics.update(predicted, target)

        # Get speed as real-time-factor (i.e., seconds decoded per second)
        seconds = penn.convert.frames_to_seconds(metrics.rpas[0].count)
        rtf = {
            key: seconds / value
            for key, value in torchutil.time.results().items()}

        # Save
        results[dataset] = metrics() | {'rtf': rtf}
    with open(torbi.EVAL_DIR / f'{torbi.CONFIG}.json', 'w') as file:
        json.dump(results, file)



import contextlib
import math
import multiprocessing as mp
import os
from typing import List, Optional, Union, Dict

import numpy as np
import torch
import torchutil

import torbi
from torbi import decode


###############################################################################
# Viterbi decoding
###############################################################################


# def from_dataloader(
#     dataloader: torch.utils.data.DataLoader,
#     output_files: Dict[
#         Union[str, bytes, os.PathLike],
#         Union[str, bytes, os.PathLike]],
#     transition: Optional[torch.Tensor] = None,
#     initial: Optional[torch.Tensor] = None,
#     log_probs: bool = False,
#     save_workers: int = 0,
#     gpu: Optional[int] = None,
#     num_threads: Optional[int] = None
# ) -> None:
#     """Decode time-varying categorical distributions from dataloader

#     Arguments
#         dataloader
#             torbi DataLoader for asynchronous loading and preprocessing
#         output_files
#             A dictionary mapping input filenames to output filenames
#         transition
#             Categorical transition matrix; defaults to uniform
#             shape=(states, states)
#         initial
#             Categorical initial distribution; defaults to uniform
#             shape=(states,)
#         log_probs
#             Whether inputs are in (natural) log space
#         save_workers
#             The number of worker threads to use for async file saving
#         gpu
#             The index of the GPU to use for inference
#         num_threads
#             The number of threads to use for parallelized decoding

#     Returns
#         indices
#             The decoded bin indices
#             shape=(batch, frames)
#     """
#     # Setup progress bar
#     progress = torchutil.iterator(
#         range(0, len(dataloader.dataset)),
#         torbi.CONFIG,
#         total=len(dataloader.dataset))

#     # Iterate over dataset
#     for observation, batch_frames, batch_chunks, input_filenames in dataloader:

#         # Decode
#         indices = from_probabilities(
#             observation=observation,
#             batch_frames=batch_frames,
#             transition=transition,
#             initial=initial,
#             log_probs=log_probs,
#             gpu=gpu,
#             num_threads=num_threads)

#         # Get output filenames
#         filenames = [output_files[file] for file in input_filenames]

#         if torbi.USE_CHUNKING:

#             # Combine chunks
#             indices = torbi.data.separate(
#                 indices=indices,
#                 batch_chunks=batch_chunks,
#                 batch_frames=batch_frames)

#             # Save
#             for indices, filename in zip(indices, filenames):
#                 save(indices.cpu().detach(), filename)

#         else:

#             # Save
#             for indices, filename, frames in zip(
#                 indices.cpu().detach(),
#                 filenames,
#                 batch_frames.cpu()
#             ):
#                 save_masked(indices, filename, frames)

#         # Increment by batch size
#         progress.update(len(input_filenames))

#     # Close progress bar
#     progress.close()


def from_probabilities(
    observation: torch.Tensor,
    batch_frames: Optional[torch.Tensor] = None,
    transition: Optional[torch.Tensor] = None,
    initial: Optional[torch.Tensor] = None,
    log_probs: bool = False,
    gpu: Optional[int] = None,
    num_threads: Optional[int] = 1
) -> torch.Tensor:
    """Decode a time-varying categorical distribution

    Arguments
        observation
            Time-varying categorical distribution
            shape=(batch, frames, states)
        batch_frames
            Number of frames in each batch item; defaults to all
            shape=(batch,)
        transition
            Categorical transition matrix; defaults to uniform
            shape=(states, states)
        initial
            Categorical initial distribution; defaults to uniform
            shape=(states,)
        log_probs
            Whether inputs are in (natural) log space
        gpu
            GPU index to use for decoding. Defaults to CPU.
        num_threads
            The number of threads to use for parallelized decoding

    Returns
        indices
            The decoded bin indices
            shape=(batch, frames)
    """
    batch, frames, states = observation.shape
    if gpu == 'mps':
        device = torch.device('mps')
    else:
        device = 'cpu' if gpu is None else f'cuda:{gpu}'

    if batch_frames is None:
        batch_frames = torch.full(
            (batch,),
            frames,
            dtype=torch.int32,
            device=device)
    batch_frames = batch_frames.to(dtype=torch.int32, device=device)

    # Default to uniform initial probabilities
    if initial is None:
        initial = torch.full(
            (states,),
            math.log((1. / states) + torch.finfo(torch.float32).tiny),
            dtype=torch.float32,
            device=device)

    # Ensure initial probabilities are in log space
    else:
        if not log_probs:
            initial = torch.log(initial)
        initial = initial.to(device)

    # Default to uniform transition probabilities
    if transition is None:
        transition = torch.full(
            (states, states),
            math.log(1. / states),
            dtype=torch.float32,
            device=device)

    # Ensure transition probabilities are in log space
    else:
        if not log_probs:
            transition = torch.log(transition)
        transition = transition.to(device)

    # Ensure observation probabilities are in log space
    if not log_probs:
        observation = torch.log(observation)
    observation = observation.to(device=device, dtype=torch.float32)

    # Add epsilon for stability
    # NOTE - may break gradients
    torch.exp_(observation)
    observation += torch.finfo(torch.float32).tiny
    torch.log_(observation)

    # Decode
    with torchutil.time.context('torbi'):
        indices = decode(
            observation,
            batch_frames,
            transition,
            initial,
            num_threads=num_threads)

    return indices


def from_file(
    input_file: Union[str, os.PathLike],
    transition_file: Optional[Union[str, os.PathLike]] = None,
    initial_file: Optional[Union[str, os.PathLike]] = None,
    log_probs: bool = False,
    gpu: Optional[int] = None,
    num_threads: Optional[int] = 1
) -> torch.Tensor:
    """Decode a time-varying categorical distribution file

    Arguments
        input_file
            Time-varying categorical distribution file
            shape=(frames, states)
        transition_file
            Categorical transition matrix file; defaults to uniform
            shape=(states, states)
        initial_file
            Categorical initial distribution file; defaults to uniform
            shape=(states,)
        log_probs
            Whether inputs are in (natural) log space
        gpu
            GPU index to use for decoding. Defaults to CPU.
        num_threads
            The number of threads to use for parallelized decoding

    Returns
        indices
            The decoded bin indices
            shape=(frames,)
    """
    # Load observations
    observation = torch.load(input_file).unsqueeze(dim=0)

    # Load transition probabilities
    if transition_file:
        transition = torch.load(transition_file)
        if log_probs:
            transition = torch.log(transition)
    else:
        transition = None

    # Load initial distribution
    if initial_file:
        initial = torch.load(initial_file)
    else:
        initial = None

    # Decode
    return from_probabilities(
        observation=observation,
        transition=transition,
        initial=initial,
        log_probs=log_probs,
        gpu=gpu,
        num_threads=num_threads)


def from_file_to_file(
    input_file: Union[str, os.PathLike],
    output_file: Union[str, os.PathLike],
    transition_file: Optional[Union[str, os.PathLike]] = None,
    initial_file: Optional[Union[str, os.PathLike]] = None,
    log_probs: bool = False,
    gpu: Optional[int] = None,
    num_threads: Optional[int] = None
) -> None:
    """Decode a time-varying categorical distribution file and save

    Arguments
        input_file
            Time-varying categorical distribution file
            shape=(frames, states)
        output_file
            File to save decoded indices
        transition_file
            Categorical transition matrix file; defaults to uniform
            shape=(states, states)
        initial_file
            Categorical initial distribution file; defaults to uniform
            shape=(states,)
        log_probs
            Whether inputs are in (natural) log space
        gpu
            GPU index to use for decoding. Defaults to CPU.
        num_threads
            The number of threads to use for parallelized decoding
    """
    indices = from_file(
        input_file,
        transition_file,
        initial_file,
        log_probs,
        gpu=gpu,
        num_threads=num_threads)
    torch.save(indices, output_file)


def from_files_to_files(
    input_files: List[Union[str, os.PathLike]],
    output_files: List[Union[str, os.PathLike]],
    transition_file: Optional[Union[str, os.PathLike]] = None,
    initial_file: Optional[Union[str, os.PathLike]] = None,
    log_probs: bool = False,
    gpu: Optional[int] = None,
    num_threads: Optional[int] = None
) -> None:
    """Decode time-varying categorical distribution files and save

    Arguments
        input_files
            Time-varying categorical distribution files
            shape=(frames, states)
        output_files
            Files to save decoded indices
        transition_file
            Categorical transition matrix file; defaults to uniform
            shape=(states, states)
        initial_file
            Categorical initial distribution file; defaults to uniform
            shape=(states,)
        log_probs
            Whether inputs are in (natural) log space
        gpu
            GPU index to use for decoding. Defaults to CPU.
        num_threads
            The number of threads to use for parallelized decoding
    """
    # Load transition probabilities
    if transition_file:
        transition = torch.load(transition_file)
        if log_probs:
            transition = torch.log(
                transition + torch.finfo(transition.dtype).tiny)
    else:
        transition = None

    # Load initial distribution
    if initial_file:
        initial = torch.load(initial_file)
    else:
        initial = None

    # Preserve file mapping
    mapping = {
        input_file: output_file
        for input_file, output_file in zip(input_files, output_files)}

    # Decode
    from_dataloader(
        dataloader=torbi.data.loader(input_files),
        output_files=mapping,
        transition=transition,
        initial=initial,
        log_probs=log_probs,
        gpu=gpu,
        num_threads=num_threads)


###############################################################################
# Utilities
###############################################################################


def from_dataloader(
    dataloader: torch.utils.data.DataLoader,
    output_files: Dict[
        Union[str, bytes, os.PathLike],
        Union[str, bytes, os.PathLike]],
    transition: Optional[torch.Tensor] = None,
    initial: Optional[torch.Tensor] = None,
    log_probs: bool = False,
    gpu: Optional[int] = None,
    num_threads: Optional[int] = 1
) -> None:
    """Decode time-varying categorical distributions from dataloader

    Arguments
        dataloader
            torbi DataLoader object for asynchronous loading and preprocessing
        output_files
            A dictionary mapping input filenames to output filenames
        transition
            Categorical transition matrix; defaults to uniform
            shape=(states, states)
        initial
            Categorical initial distribution; defaults to uniform
            shape=(states,)
        log_probs
            Whether inputs are in (natural) log space
        gpu
            The index of the GPU to use for inference

    Returns
        indices
            The decoded bin indices
            shape=(batch, frames)
    """
    # Setup progress bar
    progress = torchutil.iterator(
        range(0, len(dataloader.dataset)),
        torbi.CONFIG,
        total=len(dataloader.dataset))

    # Iterate over dataset
    for (
        observation,
        batch_frames,
        batch_chunks,
        input_filenames
    ) in dataloader:

        # Decode
        indices = from_probabilities(
            observation=observation,
            batch_frames=batch_frames,
            transition=transition,
            initial=initial,
            log_probs=log_probs,
            gpu=gpu,
            num_threads=num_threads)

        # Get output filenames
        filenames = [output_files[file] for file in input_filenames]

        # Save
        if torbi.MIN_CHUNK_SIZE is not None:

            # Combine chunks
            indices = torbi.data.separate(
                indices=indices,
                batch_chunks=batch_chunks,
                batch_frames=batch_frames)

            # Save
            for indices, filename in zip(indices, filenames):
                save(indices.cpu().detach(), filename)
        else:

            # Save
            for indices, filename, frames in zip(
                indices.cpu().detach(),
                filenames,
                batch_frames.cpu()
            ):
                save_masked(indices, filename, frames)

        # Increment by batch size
        progress.update(len(input_filenames))

    # Close progress bar
    progress.close()


def save(tensor, file):
    """Save tensor"""
    torch.save(tensor.clone(), file)


def save_masked(tensor, file, length):
    """Save masked tensor"""
    torch.save(tensor[..., :length].clone(), file)



import os
from setuptools import find_packages, setup
import numpy as np
import torch
from torch.utils.cpp_extension import (
    CppExtension,
    CUDAExtension,
    BuildExtension,
    CUDA_HOME,
)
import platform
from pathlib import Path
import glob
import warnings

library_name = "torbi"

if torch.__version__ >= "2.6.0":
    py_limited_api = True
else:
    py_limited_api = False

def get_extensions():
    debug_mode = os.getenv("DEBUG", "0") == "1"
    use_cuda = os.getenv("USE_CUDA", "1") == "1"
    if debug_mode:
        print("Compiling in debug mode")

    use_cuda = use_cuda and torch.cuda.is_available() and CUDA_HOME is not None
    extension = CUDAExtension if use_cuda else CppExtension

    if use_cuda:
        warnings.warn(f"Building torbi with CUDA using CUDA_HOME={CUDA_HOME}")

    windows = platform.system() == "Windows"
    macos = platform.system() == "Darwin"

    if windows:
        cxx_args = [
            '/O2',
            '/openmp'
            '/D Py_LIMITED_API=0x03090000' # min CPython version 3.9
        ]
        if debug_mode:
            raise ValueError('debug_mode not currently supported on windows')
    elif macos:
        cxx_args = [
            '-Xclang',
            '-fopenmp',
            # '-L/opt/homebrew/opt/libomp/lib',
            # '-I/opt/homebrew/opt/libomp/include',
            # '-lomp',
            "-O3" if not debug_mode else "-O0",
            "-fdiagnostics-color=always",
            "-DPy_LIMITED_API=0x03090000",  # min CPython version 3.9
        ]
    else: # linux
        cxx_args = [
            "-O3" if not debug_mode else "-O0",
            "-fdiagnostics-color=always",
            "-fopenmp",
            "-DPy_LIMITED_API=0x03090000",  # min CPython version 3.9
        ]

    extra_link_args = []
    extra_compile_args = {
        "cxx": cxx_args,
        "nvcc": [
            "-O3" if not debug_mode else "-O0",
        ],
    }
    if debug_mode:
        extra_compile_args["cxx"].append("-g")
        extra_compile_args["nvcc"].append("-g")
        extra_link_args.extend(["-O0", "-g"])

    this_dir = os.path.dirname(os.path.curdir)
    extensions_dir = os.path.join(this_dir, library_name, "csrc")
    sources = list(glob.glob(os.path.join(extensions_dir, "*.cpp")))

    extensions_cuda_dir = os.path.join(extensions_dir, "cuda")
    cuda_sources = list(glob.glob(os.path.join(extensions_cuda_dir, "*.cu")))

    if use_cuda:
        sources += cuda_sources

    ext_modules = [
        extension(
            f"{library_name}._C",
            sources,
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
            py_limited_api=py_limited_api,
        )
    ]

    return ext_modules

setup(
    ext_modules=get_extensions(),
    cmdclass={'build_ext': BuildExtension},
    include_dirs=[np.get_include(), 'torbi'],
    packages=find_packages(),
    options={"bdist_wheel": {"py_limited_api": "cp39"}} if py_limited_api else {},
)



from pathlib import Path


###############################################################################
# Metadata
###############################################################################


# Configuration name
CONFIG = 'torbi'


###############################################################################
# Directories
###############################################################################


# Root location for saving outputs
ROOT_DIR = Path(__file__).parent.parent.parent

# Location to save assets to be bundled with pip release
ASSETS_DIR = Path(__file__).parent.parent / 'assets'

# Location of preprocessed features
CACHE_DIR = ROOT_DIR / 'data' / 'cache'

# Location of datasets on disk
DATA_DIR = ROOT_DIR / 'data' / 'datasets'

# Location to save evaluation artifacts
EVAL_DIR = ROOT_DIR / 'eval'


###############################################################################
# Decoding
###############################################################################


# When set to a positive integer, enables chunking for long sequences by
# splitting sequences at low-entropy frames
MIN_CHUNK_SIZE = None

# Threshold below which to split the sequence when performing chunked decoding
ENTROPY_THRESHOLD = 0.5


###############################################################################
# Evaluation
###############################################################################


# Otherwise compare against self with no chunking
COMPARE_WITH_REFERENCE = True

# Names of all datasets
DATASETS = ['daps', 'vctk']

# Number of randomly-selected samples to evaluate
EVALUATION_SAMPLES = 8192

# Thresholds (in number of 5 cent bins) for raw pitch accuracy evaluation
PITCH_ERROR_THRESHOLDS = [0, 1, 2]

# File for caching transition matrix for pitch decoding evaluation
PITCH_TRANSITION_MATRIX = ASSETS_DIR / 'stats' / 'transition.pt'

# Audio sampling rate
SAMPLE_RATE = 16000

# Seed for all random number generators
RANDOM_SEED = 1234


###############################################################################
# Compute
###############################################################################


# Batch size
BATCH_SIZE = 512

# Number of parallel CPU workers
NUM_WORKERS = 0


###############################################################################
# Metadata
###############################################################################


# Allows config files to detect if this module is being configured
CONFIGURING = None



import yapecs
from pathlib import Path

import torbi


###############################################################################
# Evaluate
###############################################################################


def parse_args():
    """Parse command-line arguments"""
    parser = yapecs.ArgumentParser(description='Perform evaluation')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=torbi.DATASETS,
        help='The datasets to evaluate')
    parser.add_argument(
        '--gpu',
        help='The index of the gpu to use for evaluation, or "mps"')
    parser.add_argument(
        '--num_threads',
        type=int,
        default = 1,
        help='The number of threads to use for parellelized CPU decoding'
    )
    return parser.parse_args()


torbi.evaluate.datasets(**vars(parse_args()))
